#ifndef ARCCACHE_H_
#define ARCCACHE_H_

// project: POLiTe
// authors: Jan Hadrava 2004

/** \file */

#include <map>

#ifndef WIN32
#include <ext/hash_map>
#endif /*WIN32*/

#include "iopccommon/defs.h"
#include "arcCacheBase.h"
#include "arcCacheItem.h"
#include "types.h"
#include "iopccommon/condVar.h"


namespace iopc {

template<typename MapType> class ArcCache;

typedef ArcCache<std::map<PersIdentification, ArcCacheItem *> > MapArcCache;

#ifndef WIN32
typedef ArcCache<__gnu_cxx::hash_map<PersIdentification, ArcCacheItem *> > HashArcCache;
#endif /*WIN32*/

/** Multithreaded cache using ARC replacement policy. Objects are represented by ArcCacheItem.
* Items are stored an container conforming to STL "pair associative container" concept. Its type
* is specified by template parameter.
*/
template<typename MapType> class ArcCache: public ArcCacheBase
{
private:
#ifndef WIN32
	typedef class MapType::iterator Iter;	// MSVC6 doesn't like this although it's perfectly conforming to C++ standard
#else
	typedef MapType::iterator Iter;		// gcc doesn't like this
#endif /*WIN32*/
public:
	inline ArcCache(long minValue = MINLONG, long maxCost = MAXLONG, const Strategy& strat = Strategy::defaultStrategy);
	inline ArcCache(long minValue, long maxCost, StrategySelector* stratSel);
	inline virtual ~ArcCache();
private:/*hiding*/
	ArcCache(const ArcCache<MapType>& );
	ArcCache<MapType> &operator=(const ArcCache<MapType>& );
protected:
	virtual void clear(const StrategySelector& strSel);

	virtual CacheLock& makePersistent(const StrategySelector& strSel, const PersIdentification& persId, DatabaseObject& objectPtr);
	virtual CacheLock& getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly = false);

	virtual void dbDelete(const StrategySelector& strSel, const PersIdentification& persId);
	virtual void cacheDelete(const StrategySelector& strSel, const PersIdentification& persId);

	virtual void updateDirty(const StrategySelector& strSel);
	virtual void removeAll(const StrategySelector& strSel);
protected:
	void awaitNonBusy(ArcCacheItem& item);
	ArcCacheItem& addLock(ArcCacheItem& item, bool rdonly);
	void rmLock(ArcCacheItem& item);
public:
	virtual inline long getActiveCost();
	virtual inline long getLockedCost();
	virtual long getWorstValue(const TimeStamp& current);
	virtual long removeWorst();
	virtual bool maintainerStep(long& costChange);
	virtual long getSleepTime(const TimeStamp& current, long minValue);
	inline virtual Cache* clone() {
		return new ArcCache<MapType>(*this);
	}
protected:
	/** Remove one object scheduled for deletion.
	* \return false there are no objects scheduled for deletion, true otherwise.
	*/
	virtual bool removeDoomed();
	/** Removes one zombie. Zombies are created when an item can't removed
	* from container because of the danger of deadlock.
	* \return false when there are no zombies present, true otherwise.
	*/
	virtual bool removeZombie();
	/** Updates one modified unlocked object.
	* \param costChange Output parameter, the change of total cost caused by method call.
	* \return false when there are no modified unlocked objects present, true otherwise.
	*/
	virtual bool asyncUpdate(long& costChange);
protected:
	MapType cont;		/**< Container holding items of all types - locked/unlocked, representing scheduled actions, modified, history, ... */
	RWLock contLock;	/**< Locks items in container. */
	Mutex listLock;		/**< Common lock needed when modifying lists. */
	CondVar busyCond;	/**< Conditional for signaling state change from ST_BUSY to something else. */
	long rActiveCnt;	/**< Number of unlocked objects (recent). */
	long rLockedCnt;	/**< Number of locked objects (recent). */
	long fActiveCnt;	/**< Number of unlocked objects (frequent). */
	long fLockedCnt;	/**< Number of locked objects (frequent). */
	long rHistCnt;		/**< Number of items representing history (recent). */
	long fHistCnt;		/**< Number of items representing history (frequent). */
	ArcCacheItem rCleanList;/**< List of objects which doesn't require async update (recent). */
	ArcCacheItem rHistList; /**< List of history records (recent). */
	ArcCacheItem rDirtyList;/**< List of objects requiring async update (recent). */
	ArcCacheItem fCleanList;/**< List of objects which doesn't require async update (frequent). */
	ArcCacheItem fHistList; /**< List of history records (frequent). */
	ArcCacheItem fDirtyList;/**< List of objects requiring async update (frequent). */
	ArcCacheItem doomedList;/**< List of items scheduled for deletion. */
	ArcCacheItem zombieList;/**< List of zombies. */
	long activeCost;	/**< Total cost of unlocked items (recent and frequent). */
	long lockedCost;	/**< Total cost of locked items (recent and frequent). */
	double target;		/**< Target for ratio activeCost/lockedCost. */
};

template<typename MapType> ArcCache<MapType>::ArcCache(long minValue, long maxCost, StrategySelector* stratSel)
: ArcCacheBase(minValue, maxCost, stratSel),
	cont(), contLock(), listLock(), busyCond(),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rCleanList(), rHistList(),  rDirtyList(), fCleanList(), fHistList(), fDirtyList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCache<MapType>::ArcCache(long minValue, long maxCost, const Strategy& strat)
: ArcCacheBase(minValue, maxCost, strat),
	cont(), contLock(), listLock(), busyCond(),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rCleanList(), rHistList(),  rDirtyList(), fCleanList(), fHistList(), fDirtyList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCache<MapType>::ArcCache (const ArcCache<MapType>& src)
: ArcCacheBase(src.minValue, src.maxCost, src.stratSel, src.defaultStrategy),
	cont(), contLock(), listLock(), busyCond(),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rCleanList(), rHistList(),  rDirtyList(), fCleanList(), fHistList(), fDirtyList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCache<MapType>::~ArcCache()
{
	ExtendedCache::clear();
}

template<typename MapType> void ArcCache<MapType>::clear(const StrategySelector& strSel)
{

	contLock.wlock();

	long costChange = 0;
	for (Iter iter = cont.begin();  iter != cont.end(); iter++) {
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)

			contLock.unlock();	/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		long cost;
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* new item */
			if (item->getObject().isDirty()) {
				try {
					persId.insert(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)

					contLock.unlock();	/* avoiding deadlock */
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			listLock.lock();
			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			listLock.unlock();
		break;
		case ArcCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				try {
					persId.update(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)

					contLock.unlock();	/* avoiding deadlock */
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			listLock.lock();
			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			listLock.unlock();
		break;
		case ArcCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			try {
				persId.remove(*conn);
			}
			IOPC_LOG_CATCH(IopcException& exc)

				contLock.unlock();	/* avoiding deadlock */
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			listLock.lock();
			item->remove();
			listLock.unlock();
		break;
		case ArcCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			listLock.lock();
			item->remove();
			listLock.unlock();
		break;
		case ArcCacheItem::ST_HIST:
			listLock.lock();
			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;
			listLock.unlock();
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}

		delete item;
	}
	cont.clear();

//	IOPC_ASSERT(fActiveCnt == 0 && rActiveCnt == 0 && fLockedCnt == 0 && rLockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(rCleanList.prev == &rCleanList && rCleanList.next == &rCleanList);
	IOPC_ASSERT(fCleanList.prev == &fCleanList && fCleanList.next == &fCleanList);
	IOPC_ASSERT(rHistList.prev == &rHistList && rHistList.next == &rHistList);
	IOPC_ASSERT(fHistList.prev == &fHistList && fHistList.next == &fHistList);
	IOPC_ASSERT(rDirtyList.prev == &rDirtyList && rDirtyList.next == &rDirtyList);
	IOPC_ASSERT(fDirtyList.prev == &fDirtyList && fDirtyList.next == &fDirtyList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == &doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == &zombieList);
	const Strategy& strat = strSel.select();

	contLock.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> CacheLock &ArcCache<MapType>::makePersistent(const StrategySelector& strSel, const PersIdentification& persId, class DatabaseObject &objectPtr)
{
	long costChange;

	contLock.wlock();

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:

			contLock.unlock();
			IOPC_LOG_THROW(IopcDuplicateKeyException());
		default:
			IOPC_UNREACHABLE;
		break;
		}
		ArcCacheItem::State oldstate = item->state;
		item->state = ArcCacheItem::ST_BUSY;	/* we will keep the item in the cache */

		contLock.unlock();

		switch (oldstate) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:
			IOPC_UNREACHABLE;
		break;
		case ArcCacheItem::ST_HIST:
			/* this is strange - item with the same ID as the one being inserted was recently evicted from cache */
			listLock.lock();
			item->remove();	/* remove from histList */
			item->isFrequent() ? fHistCnt-- : rHistCnt--;
			listLock.unlock();
		break;
		case ArcCacheItem::ST_DOOMED:
			try {
				persId.remove(*conn);	/* doing maintainer's work */
			}
			IOPC_LOG_CATCH(IopcException& exc)
				busyCond.lock();
				item->state = ArcCacheItem::ST_DOOMED;
				busyCond.signal();
				busyCond.unlock();
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
		/*nobreak;*/
		case ArcCacheItem::ST_ZOMBIE:
			listLock.lock();
			item->remove();
			listLock.unlock();
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		/* now we have in cache busy item which is prepared to become born/adult */
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		cont[persId] = item;

		contLock.unlock();

	}
	IOPC_ASSERT(item->state == ArcCacheItem::ST_BUSY);
	UpdatingStrategy uStrat = strat.getUpdatingStrategy();
	if (uStrat == USImmed && strat.getLoadLocking()) {
		uStrat = USDeferred;	/* it's not possible to immediately insert object which should be locked in database */
	}
	if (uStrat == USImmed) {
		objectPtr.setDirty(false);
	} else {
		objectPtr.setDirty(true);
	}
	if (uStrat == USImmed) {
		try {
			persId.insert(*conn, objectPtr);
		}
		IOPC_LOG_CATCH(IopcException& exc)
			listLock.lock();
			zombieList.addPrev(*item);
			listLock.unlock();
			costChange = 0;
			busyCond.lock();
			item->state = ArcCacheItem::ST_ZOMBIE;
			busyCond.signal();
			busyCond.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		objectPtr.setDirty(false);
	}
	item->isLocked = true;
	item->setFrequent(false);
	item->addLock(false);
	//objectPtr.SetPersistent(*conn);
	item->objectPtr = &objectPtr;
	busyCond.lock();
	item->state = uStrat == USImmed ? ArcCacheItem::ST_ADULT : ArcCacheItem::ST_BORN;
	busyCond.signal();
	busyCond.unlock();

	listLock.lock();
	rLockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());
	listLock.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> CacheLock &ArcCache<MapType>::getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly)
{

	contLock.wlock();

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT: {
		/* item is present and active */
			item->addLock(rdonly);
			listLock.lock();
			if (item->isLocked) {
				/* item was already locked,
				** so there's no need to remove it to cleanList
				** and it will be unsafe to reload it */
				listLock.unlock();

				contLock.unlock();

				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* first lock added, item must be added to lockedList and can be reloaded */
			item->remove(); /* removes item from activeList or updateList */
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost -= strat.cost(item->getObject());
			fLockedCnt++;
			lockedCost += strat.cost(item->getObject());
			item->setFrequent(true);
			item->isLocked = true;
			listLock.unlock();
			ReadingStrategy rStrat = strat.getReadingStrategy();
			if (item->state == ArcCacheItem::ST_BORN || rStrat == RSCache || item->getObject().isDirty()) {
				/* there's no need to reload object */

				contLock.unlock();

				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* object will be reloaded */
			IOPC_ASSERT(item->state == ArcCacheItem::ST_ADULT);
			item->state = ArcCacheItem::ST_BUSY;

			contLock.unlock(); /* avoid blocking whole cache during reloading */

			try {
				switch (rStrat) {
				case RSTimestamp:
					if (persId.versionMatches(*conn, dynamic_cast<OidObject&>(item->getObject()))) {
						break;
					}
				/*nobreak*/
				case RSDatabase: {
					DatabaseObject* objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
					IOPC_ASSERT(objectPtr != NULL);
					item->replaceObject(objectPtr);
				break; }
				case RSCache:
				break;
				default:
					IOPC_UNREACHABLE;
				break;
				}
			}
			IOPC_LOG_CATCH(IopcException&)	/* reload failed, ignore it */
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			busyCond.lock();
			item->state = ArcCacheItem::ST_ADULT;
			busyCond.signal();
			busyCond.unlock();
			strat.wakeUpMaintainer(0);
			imposeLimits();
			return *item;
		break; }
		case ArcCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */

			contLock.unlock();
			IOPC_LOG_THROW(IopcNotFoundException("Item not present"));
		break;
		case ArcCacheItem::ST_ZOMBIE:
			listLock.lock();
			item->remove();
			listLock.unlock();
			item->setFrequent(false);
		break;
		case ArcCacheItem::ST_HIST:
			listLock.lock();
			/* adaptation */
			if (item->isFrequent()) {
				IOPC_ASSERT(fHistCnt > 0);
				long fCnt = fActiveCnt + fLockedCnt;
				if (fCnt == 0) {
					fCnt = 1;
				}
				target -= ((0.0 + rHistCnt)/fHistCnt)/(fCnt);
				if (target < 0.0) {
					target = 0.0;
				}
			} else {
				IOPC_ASSERT(rHistCnt > 0);
				long fCnt = fActiveCnt + fLockedCnt;
				if (fCnt == 0) {
					fCnt = 1;
				}
				target += ((0.0 + fHistCnt)/rHistCnt)/(fCnt);
				if (target > 1.0) {
					target = 1.0;
				}
			}
			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;
			listLock.unlock();
			item->setFrequent(true);
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		item->state = ArcCacheItem::ST_BUSY;
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		item->setFrequent(false);
		cont[persId] = item;
	}

	contLock.unlock(); /* item added to cache, contLock can be unlocked */

	DatabaseObject* objectPtr;
	try {
		objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
	}
	IOPC_LOG_CATCH(IopcException& exc)
		listLock.lock();
		zombieList.addPrev(*item);
		listLock.unlock();
		busyCond.lock();
		item->state = ArcCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	IOPC_ASSERT(objectPtr != NULL);
	//objectPtr->SetPersistent(*conn);
	item->objectPtr = objectPtr;
	item->addLock(rdonly);
	item->isLocked = true;
	listLock.lock();
	long costChange;
	item->isFrequent() ? fLockedCnt++ : rLockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());
	listLock.unlock();
	busyCond.lock();
	item->state = ArcCacheItem::ST_ADULT;
	busyCond.signal();
	busyCond.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> void ArcCache<MapType>::dbDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	long costChange = 0;

	contLock.wlock();

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)

			contLock.unlock();	/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* remove it immediately */
			listLock.lock();
			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
			listLock.unlock();
			cont.erase(iter);

			contLock.unlock();

			delete item;
			strat.wakeUpMaintainer(costChange);
			imposeLimits();
			return; /* ignoring the possibility an item already present in the database
				** can have the same key as the removed one */
		break;
		case ArcCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */

			contLock.unlock();
			IOPC_LOG_THROW(IopcNotFoundException("Item not present"));
		break;
		case ArcCacheItem::ST_ADULT:
			listLock.lock();
			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
			listLock.unlock();
			item->release();
			item->state = ArcCacheItem::ST_BUSY;

			contLock.unlock();

		break;
		case ArcCacheItem::ST_ZOMBIE:
			listLock.lock();
			item->remove();
			item->state = ArcCacheItem::ST_BUSY;
			listLock.unlock();
			item->release();	/* release data */

			contLock.unlock();

			costChange = 0;
		break;
		case ArcCacheItem::ST_HIST:
			listLock.lock();
			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;
			item->state = ArcCacheItem::ST_BUSY;
			listLock.unlock();
			item->release();	/* release data */

			contLock.unlock();

			costChange = 0;
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		//IOPC_UNREACHABLE;
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		cont[persId] = item;

		contLock.unlock();

		costChange = 0;
	}
	IOPC_ASSERT(item->state == ArcCacheItem::ST_BUSY);
	if (strat.getUpdatingStrategy() == USImmed) {
		listLock.lock();
		zombieList.addPrev(*item);
		listLock.unlock();
		try {
			persId.remove(*conn);
		}
		IOPC_LOG_CATCH(IopcException& exc)
			busyCond.lock();
			item->state = ArcCacheItem::ST_ZOMBIE;
			busyCond.signal();
			busyCond.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		busyCond.lock();
		item->state = ArcCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
	} else { /* USAsync || USImmedDeferred */
		listLock.lock();
		doomedList.addPrev(*item);	/* schedule for deletion */
		listLock.unlock();
		busyCond.lock();
		item->state = ArcCacheItem::ST_DOOMED;
		busyCond.signal();
		busyCond.unlock();
	}
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void ArcCache<MapType>::cacheDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	/* ignoring user's advices */
}

template<typename MapType> void ArcCache<MapType>::updateDirty(const StrategySelector& strSel)
{
	Iter next;

	contLock.wlock();

	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter = next) {
		next = iter;
		next++;
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)

			contLock.unlock();	/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* new item */
			IOPC_ASSERT(item->getObject().isDirty());
			persId.insert(*conn, item->getObject());
			item->getObject().setDirty(false);
			item->state = ArcCacheItem::ST_ADULT;
		break;
		case ArcCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				persId.update(*conn, item->getObject());
				item->getObject().setDirty(false);
			}
		break;
		case ArcCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			persId.remove(*conn);
			listLock.lock();
			item->remove();
			listLock.unlock();
			cont.erase(iter);
			delete item;
		break;
		case ArcCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			listLock.lock();
			item->remove();
			listLock.unlock();
			cont.erase(iter);
			delete item;
		break;
		case ArcCacheItem::ST_HIST:
			/* keep history */
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		if ((item->state == ArcCacheItem::ST_BORN || item->state == ArcCacheItem::ST_ADULT)
                        && strat.getLoadLocking())
                {       /* item with locking must be removed */
			listLock.lock();
			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			long cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			listLock.unlock();
			delete item;
			cont.erase(iter);
		}
	}
	IOPC_ASSERT(doomedList.next == &doomedList && doomedList.prev == &doomedList);
	IOPC_ASSERT(zombieList.next == &zombieList && zombieList.prev == &zombieList);


	contLock.unlock();

	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void ArcCache<MapType>::removeAll(const StrategySelector& strSel)
{
	long costChange = 0;

	contLock.wlock();

	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		awaitNonBusy(*item);
		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)

			contLock.unlock();	/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		listLock.lock();
		item->remove();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
		break;
		case ArcCacheItem::ST_DOOMED:
		case ArcCacheItem::ST_ZOMBIE:
		case ArcCacheItem::ST_HIST:
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		listLock.unlock();
		delete item;
	}
	cont.clear();
	IOPC_ASSERT(rActiveCnt == 0 && rLockedCnt == 0);
	IOPC_ASSERT(fActiveCnt == 0 && fLockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(rCleanList.prev == &rCleanList && rCleanList.next == & rCleanList);
	IOPC_ASSERT(fCleanList.prev == &fCleanList && fCleanList.next == & fCleanList);
	IOPC_ASSERT(rDirtyList.prev == &rDirtyList && rDirtyList.next == & rDirtyList);
	IOPC_ASSERT(fDirtyList.prev == &fDirtyList && fDirtyList.next == & fDirtyList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == & doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == & zombieList);


	contLock.unlock();

	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void ArcCache<MapType>::awaitNonBusy(ArcCacheItem& item)
{
	busyCond.lock();
	while (item.state == ArcCacheItem::ST_BUSY) {
		busyCond.wait();
	}
	busyCond.unlock();
}

template<typename MapType> ArcCacheItem &ArcCache<MapType>::addLock(ArcCacheItem& item, bool rdonly)
{
	if (rdonly) {
		item.lock.rlock();
	} else {
		item.lock.wlock();
	}
	return item;
}

template<typename MapType> void ArcCache<MapType>::rmLock(ArcCacheItem& item)
{
	/* it's important not to use contLock here so that we avoid deadlock */
	/* it's important not to use contLock here so that we avoid deadlock */
	try {
		item.lock.unlock();
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	listLock.lock();
	if (!item.isLocked) {
		listLock.unlock();
		return;
	}
	if (!item.lock.tryXLock()) {
		listLock.unlock();
		return;
	}
	item.remove();	/* remove from list of locked items */
	long cost = item.strat.cost(item.getObject());
	long costChange = 0;
	if (item.isFrequent()) {
		fLockedCnt--;
		fActiveCnt++;
	} else {
		rLockedCnt--;
		rActiveCnt++;
	}
	lockedCost -= cost;
	activeCost += cost;
	item.isLocked = false;
	UpdatingStrategy uStrat = item.strat.getUpdatingStrategy();
	if (uStrat == USImmed && item.state == ArcCacheItem::ST_BORN) {
		IOPC_ASSERT(item.strat.getLoadLocking());
		/* this combination will require inserting item into db and removing local copy but we can't do it
		because it requires state change which can't be made without contLock */
		uStrat = USAsync;
	}
	switch (uStrat) {
	case USDeferred:
		item.isFrequent() ? fCleanList.addNext(item) : rCleanList.addNext(item);
	break;
	case USImmed:
		if (item.state == ArcCacheItem::ST_BORN) {
			IOPC_UNREACHABLE;
		} else {
			IOPC_ASSERT(item.state == ArcCacheItem::ST_ADULT);
			if (item.getObject().isDirty()) {
				listLock.unlock();	/* this is safe because we have xlock for the item */
				item.persId.update(*conn, item.getObject());
				item.getObject().setDirty(false);
				listLock.lock();
			}
			item.isFrequent() ? fCleanList.addNext(item) : rCleanList.addNext(item);
		}
	break;
	case USAsync:
		if (item.getObject().isDirty()) {
			item.isFrequent() ? fDirtyList.addNext(item) : rDirtyList.addNext(item);
		} else {
			item.isFrequent() ? fCleanList.addNext(item) : rCleanList.addNext(item);
		}
	break;
	}
	item.timeStamp = TimeStamp::now();
	Strategy strat = item.strat;
	item.lock.unlock();
	listLock.unlock();
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> inline long ArcCache<MapType>::getActiveCost()
{
	return activeCost;
}

template<typename MapType> inline long ArcCache<MapType>::getLockedCost()
{
	return lockedCost;
}

template<typename MapType> long ArcCache<MapType>::getWorstValue(const TimeStamp& time)
{
	long retval = MAXLONG;
	listLock.lock();
	if (rCleanList.prev != &rCleanList) {
		retval = rCleanList.prev->timeStamp - time;	/* difference in milliseconds */
	}
	if (rDirtyList.prev != &rDirtyList) {
		long value = rDirtyList.prev->timeStamp - time;
		if (value < retval) {
			retval = value;
		}
	}
	if (fCleanList.prev != &fCleanList) {
		long value = fCleanList.prev->timeStamp - time;
		if (value < retval) {
			retval = value;
		}
	}
	if (fDirtyList.prev != &fDirtyList) {
		long value = fDirtyList.prev->timeStamp - time;
		if (value < retval) {
			retval = value;
		}
	}
	listLock.unlock();
	return retval;
}

template<typename MapType> long ArcCache<MapType>::removeWorst()
{

	contLock.wlock();

	listLock.lock();
	//ArcCacheItem* list = NULL;
	long value;
	TimeStamp now = TimeStamp::now();
	ArcCacheItem* item = NULL;
	ArcCacheItem* cleanList, *dirtyList;
	bool freq;
	long cost = 0;
	if (fActiveCnt + fLockedCnt == 0 || (rActiveCnt + rLockedCnt)/(fActiveCnt + fLockedCnt) > target) {
		freq = false;
		cleanList = &rCleanList;
		dirtyList = &rDirtyList;
	} else {
		freq = true;
		cleanList = &fCleanList;
		dirtyList = &fDirtyList;
	}
	if (cleanList->prev != cleanList) {
		value = cleanList->prev->timeStamp - now;
		item = cleanList->prev;
	}
	if (dirtyList->prev != dirtyList) {
		if (item == NULL || value > dirtyList->prev->timeStamp - now) {
			item = dirtyList->prev;
		}
	}
	if (item == NULL) {
		listLock.unlock();

		contLock.unlock();

	} else {
		awaitNonBusy(*item);
		if (item->state == ArcCacheItem::ST_BUSY) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Incorrect behavior of maintainer"));
		}
		bool isAdult;
		if (item->state == ArcCacheItem::ST_ADULT) {
			isAdult = true;
		} else {
			IOPC_ASSERT(item->state == ArcCacheItem::ST_BORN);
			isAdult = false;
		}
		item->state = ArcCacheItem::ST_BUSY;
		item->remove();
		//cost;
		activeCost += cost = -item->strat.cost(item->getObject());
		if (freq) {
			fActiveCnt--;
			fHistCnt++;
			fHistList.addNext(*item);
		} else {
			rActiveCnt--;
			rHistCnt++;
			rHistList.addNext(*item);
		}
		listLock.unlock();

		contLock.unlock();

		if (item->getObject().isDirty()) {
			if (isAdult) {
				item->persId.update(*conn, item->getObject());
			} else {
				item->persId.insert(*conn, item->getObject());
			}
			item->getObject().setDirty(false);
		}
		item->release();
		busyCond.lock();
		item->state = ArcCacheItem::ST_HIST;
		busyCond.signal();
		busyCond.unlock();
	}


	contLock.wlock();

	listLock.lock();


	while (fHistCnt > rActiveCnt + rLockedCnt && fHistList.prev != &fHistList) {

		item = fHistList.prev;

		item->remove();

		item->release();

		Iter iter = cont.find(item->persId);

		IOPC_ASSERT(iter != cont.end());

		cont.erase(iter);

		delete item;

		fHistCnt--;
	}


	while (rHistCnt > fActiveCnt + fLockedCnt && rHistList.prev != &rHistList) {

		item = rHistList.prev;
		item->remove();
		item->release();
		Iter iter = cont.find(item->persId);
		IOPC_ASSERT(iter != cont.end());
		cont.erase(iter);
		delete item;
		rHistCnt--;
	}

	listLock.unlock();

	contLock.unlock();


	return cost;
}

template<typename MapType> bool ArcCache<MapType>::removeDoomed()
{

	contLock.wlock();

	listLock.lock();
	if (doomedList.prev == &doomedList) {
		listLock.unlock();

		contLock.unlock();

		return false;
	}
	ArcCacheItem& item = *doomedList.prev;
	awaitNonBusy(item);
	IOPC_ASSERT(item.state == ArcCacheItem::ST_DOOMED);
	item.state = ArcCacheItem::ST_BUSY;
	item.remove();
	zombieList.addPrev(item);
	listLock.unlock();

	contLock.unlock();

	item.persId.remove(*conn);
	busyCond.lock();
	item.state = ArcCacheItem::ST_ZOMBIE;
	busyCond.signal();
	busyCond.unlock();
	return true;
}

template<typename MapType> bool ArcCache<MapType>::removeZombie()
{

	contLock.wlock();

	listLock.lock();
	if (zombieList.next == &zombieList) {
		listLock.unlock();

		contLock.unlock();

		return false;
	}
	ArcCacheItem* item = zombieList.next;
	awaitNonBusy(*item);
	IOPC_ASSERT(item->state == ArcCacheItem::ST_ZOMBIE);
	Iter iter = cont.find(item->persId);
	IOPC_ASSERT(iter != cont.end());
	cont.erase(iter);
	item->remove();
	listLock.unlock();

	contLock.unlock();

	delete item;
	return true;
}

template<typename MapType> bool ArcCache<MapType>::asyncUpdate(long& costChange)
{
	static bool freq = false;
	costChange = 0;

	contLock.wlock();

	listLock.lock();
	ArcCacheItem* dirtyList = &rDirtyList;
	if (dirtyList->prev == dirtyList || freq) {
		dirtyList = &fDirtyList;
	}
	freq = !freq;
	if (dirtyList->prev == dirtyList) {
		listLock.unlock();

		contLock.unlock();

		return false;
	}
	ArcCacheItem& item = *dirtyList->prev;
	ArcCacheItem::State oldState = item.state;
	IOPC_ASSERT(oldState == ArcCacheItem::ST_BORN || oldState == ArcCacheItem::ST_ADULT);
	item.state = ArcCacheItem::ST_BUSY;
	item.remove();
	if (oldState == ArcCacheItem::ST_BORN && item.strat.getLoadLocking()) {
		/* insert into database and remove from cache */
		zombieList.addPrev(item);
		item.isFrequent() ? fActiveCnt-- : rActiveCnt--;
		costChange = -item.strat.cost(item.getObject());
		activeCost += costChange;
		listLock.unlock();

		contLock.unlock();

		item.persId.insert(*conn, item.getObject());
		item.release();
		busyCond.lock();
		item.state = ArcCacheItem::ST_ZOMBIE;
		busyCond.signal();
		busyCond.unlock();
		return true;
	}
	ArcCacheItem* listItem;
	ArcCacheItem* cleanList = freq ? &fCleanList : &rCleanList;
	for (listItem = cleanList->next; listItem != cleanList; listItem = listItem->next) {
		if (listItem->timeStamp <= item.timeStamp) {
			break;
		}
	}
	listItem->addPrev(item);
	IOPC_ASSERT(item.prev == cleanList || item.prev->timeStamp >= item.timeStamp);
	IOPC_ASSERT(item.next == cleanList || item.next->timeStamp <= item.timeStamp);
	listLock.unlock();

	contLock.unlock();

	if (oldState == ArcCacheItem::ST_ADULT) {
	try {
		item.persId.update(*conn, item.getObject());
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CAUGHT
	} else {
		item.persId.insert(*conn, item.getObject());
	}
	item.getObject().setDirty(false);
	busyCond.lock();
	item.state = ArcCacheItem::ST_ADULT;
	busyCond.signal();
	busyCond.unlock();
	return true;
}

template<typename MapType> bool ArcCache<MapType>::maintainerStep(long& costChange)
{bool retval;
	try {
	costChange = 0;
	retval = asyncUpdate(costChange);
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CAUGHT
	if (removeDoomed()) {
		retval = true;
	}
	if (removeZombie()) {
		retval = true;
	}
	return retval;
}

template<typename MapType> long ArcCache<MapType>::getSleepTime(const TimeStamp& current, long minValue)
{
	listLock.lock();
	ArcCacheItem* item = NULL;
	long diff;

	if (rCleanList.prev != &rCleanList) {
		diff = rCleanList.prev->timeStamp - current;
		item = rCleanList.prev;
	}
	if (fCleanList.prev != &fCleanList) {
		long d = fCleanList.prev->timeStamp - current;
		if (item == NULL || diff > d) {
			diff = d;
			item = fCleanList.prev;
		}
	}
	if (rDirtyList.prev != &rDirtyList) {
		long d = rDirtyList.prev->timeStamp - current;
		if (item == NULL || diff > d) {
			diff = d;
			item = rDirtyList.prev;
		}
	}
	if (fDirtyList.prev != &fDirtyList) {
		long d = fDirtyList.prev->timeStamp - current;
		if (item == NULL || diff > d) {
			diff = d;
			item = fDirtyList.prev;
		}
	}
	if (item == NULL) {
		listLock.unlock();
///		return 1800*1000000;	/* something like infinity */
		return 5*1000000;
	}
	listLock.unlock();
	if (diff < minValue) {
		return 0;		/* no time to sleep */
	}
	return diff - minValue;
}

}/*namespace iopc*/


#endif /*ARCCACHE_H_*/
