#include "simpleStrategySelector.h"
#include "simpleCacheSelector.h"
#include "cacheKeeper.h"
#include "iopcdb/connection.h"
#include "fakeExtendedCache.h"

namespace iopc {


CacheKeeper::CacheKeeper(long minValue, long maxCost, long costLimit, bool memCost)
: ckMap(), change(false), changeCond(), mntrLock(), lock(),
	maxCost(maxCost), minValue(minValue), costLimit(costLimit), memCost(memCost),
	totalCost(0), mntrRunning(false)
{
}

CacheKeeper::~CacheKeeper()
{
	IOPC_FUNC_INVOKED;
	stopMaintainer();
	for (Iter iter = ckMap.begin(); iter != ckMap.end(); iter++) {
		CKRecord *rec = iter->second;
		rec->stratSel->offline();
		// All will be deleted by cachedConnection or composedCache */
		delete rec;
	}
	ckMap.clear();
	IOPC_FUNC_ENDED;
}

ComposedCache* CacheKeeper::setCache(CachedConnection* conn, const Type& type, Cache* cache)
{
	return setCacheInternal(conn, &type, cache);
}

ComposedCache* CacheKeeper::setCache(CachedConnection* conn, Cache* cache)
{
	return setCacheInternal(conn, NULL, cache);
}

ComposedCache* CacheKeeper::setCacheInternal(CachedConnection* conn, const Type* type, Cache* cache)
{
	if (typeid(conn->getCache()) != typeid(ComposedCache)) {
		IOPC_LOG_THROW(IopcNotSupportedException("Only cached connections with associated ComposedCache are allowed"));
	}
	lock.wlock();
	Iter iter = ckMap.find(conn);
	CKRecord* rec;
	if (iter == ckMap.end()) {
		rec = new CKRecord();
		rec->cacheSel = new SimpleCacheSelector();
		rec->stratSel = new SimpleStrategySelector((void *)this, &wakeUpMntr, memCost ? &Strategy::memoryCost : &Strategy::constCost);
		rec->cache = (ComposedCache*) &conn->getCache();
		rec->cache->setCacheSelector(rec->cacheSel);
		rec->cache->setStrategySelector(rec->stratSel);
		ckMap[conn] = rec;
		//conn.setCache(*rec->cache);
	} else {
		rec = iter->second;
	}
	ExtendedCache* advCache = NULL;
	if (cache != NULL) {
		//cache->setConnection(&conn);
		advCache = dynamic_cast<ExtendedCache *>(cache);
		if (advCache == NULL) {
			advCache = new FakeExtendedCache(cache, true); // is owned by the FakeExtendedCache
		}
	}
	/* indirectly changing respective composed cache */
	conn->stopCache();
	if (type != NULL) {
		rec->cacheSel->setCache(*type, advCache, true);
	} else {
		rec->cacheSel->setCache(advCache, true);
	}
	conn->restartCache();
	lock.unlock();
	return rec->cache;
}

ComposedCache* CacheKeeper::setStrategy(CachedConnection* conn, const Type& type, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat)
{
	return setStrategyInternal(conn, &type, loadWait, loadLocking, uStrat, rStrat);
}

ComposedCache* CacheKeeper::setStrategy(CachedConnection* conn, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat)
{
	return setStrategyInternal(conn, NULL, loadWait, loadLocking, uStrat, rStrat);
}

ComposedCache* CacheKeeper::setStrategy(CachedConnection* conn, const Type& type, const Strategy& strat)
{
	return setStrategyInternal(conn, &type, strat.getLoadWait(), strat.getLoadLocking(),
		strat.getUpdatingStrategy(), strat.getReadingStrategy());
}

ComposedCache* CacheKeeper::setStrategy(CachedConnection* conn, const Strategy& strat)
{
	return setStrategyInternal(conn, NULL, strat.getLoadWait(), strat.getLoadLocking(),
		strat.getUpdatingStrategy(), strat.getReadingStrategy());
}

ComposedCache* CacheKeeper::setStrategyInternal(CachedConnection* conn, const Type* type, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat)
{
	if (typeid(conn->getCache()) != typeid(ComposedCache)) {
		IOPC_LOG_THROW(IopcNotSupportedException("Only cached connections with associated ComposedCache are allowed"));
	}
	lock.wlock();
	Iter iter = ckMap.find(conn);
	CKRecord *rec;
	if (iter == ckMap.end()) {
		rec = new CKRecord;
		rec->cacheSel = new SimpleCacheSelector();
		rec->stratSel = new SimpleStrategySelector((void *)this, &wakeUpMntr, memCost ? &Strategy::memoryCost : &Strategy::constCost);
		rec->cache = (ComposedCache*) &conn->getCache();
		rec->cache->setCacheSelector(rec->cacheSel);
		rec->cache->setStrategySelector(rec->stratSel);
		//rec->cache = new ComposedCache(*rec->cacheSel, *rec->stratSel);
		ckMap[conn] = rec;
		//conn.SetCache(*rec->cache);
	} else {
		rec = iter->second;
	}
	/* indirectly changing respective composed cache */
	conn->stopCache();
	if (type != NULL) {
		rec->stratSel->setStrategy(*type, loadWait, loadLocking, uStrat, rStrat);
	} else {
		rec->stratSel->setStrategy(loadWait, loadLocking, uStrat, rStrat);
	}
	conn->restartCache();
	lock.unlock();
	return rec->cache;
}

void CacheKeeper::releaseConnection(CachedConnection* conn)
{
	IOPC_FUNC_INVOKED;
	lock.wlock();
	Iter iter = ckMap.find(conn);
	if (iter == ckMap.end()) {
		CKRecord* rec = iter->second;
		ckMap.erase(iter);
		//conn.setCache();
		delete rec->cache;
		delete rec->stratSel;
		delete rec->cacheSel;
		delete rec;
	}
	lock.unlock();
}

bool CacheKeeper::runMaintainer()
{
	IOPC_FUNC_INVOKED;
	lock.wlock();
	if (mntrRunning) {
		lock.unlock();
		return false;
	}
	stopMntr = false;

	pthread_attr_t tattr;
	//pthread_t tid;

	/* initialized with default attributes */
	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr,PTHREAD_CREATE_JOINABLE);


	if (pthread_create(&mntrId, &tattr, &maintainer, (void *)this) != 0) {
		lock.unlock();
		return false;
	}
	mntrRunning = true;
	lock.unlock();
	return true;
}

bool CacheKeeper::stopMaintainer()
{
	IOPC_FUNC_INVOKED;
	lock.wlock();
	if (!mntrRunning) {
		lock.unlock();
		return false;
	}
	changeCond.lock();
	change = stopMntr = true;
	changeCond.signal();
	changeCond.unlock();
	lock.unlock();
	IOPC_DEBUGLOG("pthread_join");
	bool retval = pthread_join(mntrId, NULL) == 0;
	mntrRunning = false;
	IOPC_FUNC_ENDED;
	return retval;
}

bool CacheKeeper::maintainerStep(const TimeStamp& now, long& sleepTime, long limit)
{
	IOPC_FUNC_INVOKED;
	lock.wlock();

	changeCond.lock();
	change = false;
	changeCond.unlock();

	ExtendedCache* worst = NULL;
	long worstValue = MAXLONG;
	/*long activeCost = 0;
	long lockedCost = 0;*/
	sleepTime = MAXLONG;
	bool success = false;
	long costChange = 0;
	for (Iter iter = ckMap.begin(); iter != ckMap.end(); iter++) {
		CKRecord* ckrec = iter->second;
		for (SimpleCacheSelector::Iter it = ckrec->cacheSel->cacheMap.begin(); it != ckrec->cacheSel->cacheMap.end(); it++) {
			ExtendedCache* cache = it->second->cache;
			while (cache->getWorstValue(now) < minValue) {
				long costChange = cache->removeWorst();
				if (costChange == 0) {
					break;
				}
				success = true;
				changeCond.lock();
				totalCost += costChange;
				changeCond.unlock();
			}
			while (cache->maintainerStep(costChange)) {
				changeCond.lock();
				totalCost += costChange;
				changeCond.unlock();
				success = true;
			}
			long st = cache->getSleepTime(now, minValue);
			if (st < sleepTime) {
				sleepTime = st;
			}
			long wv = cache->getWorstValue(now);
			if (wv < worstValue) {
				worstValue = wv;
				worst = cache;
			}
		}
	}
	if (totalCost > limit) {
		if (worst != NULL) {
			costChange = worst->removeWorst();
			if (costChange != 0) {
				changeCond.lock();
				totalCost += costChange;
				changeCond.unlock();
				success = true;
			}
		}
	}
	lock.unlock();
	if (success) {
		return true;
		//IOPC_LOG_RETURN(true);
	} else {
		return false;
		//IOPC_LOG_RETURN(false);
	}
}

void* CacheKeeper::maintainer(void *instance)
{
	IOPC_FUNC_INVOKED;
	CacheKeeper* ck = (CacheKeeper *)instance;
	if (ck == NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	}
	long sleepTime = 0;
	while (true) {
		ck->mntrLock.lock();
		TimeStamp now = TimeStamp::now();
		ck->maintainerStep(now, sleepTime, ck->maxCost);
		ck->mntrLock.unlock();
		ck->changeCond.lock();
		if (ck->stopMntr) {
			ck->changeCond.unlock();
			pthread_exit(NULL);
			//LOG_RETURN(NULL);
			IOPC_FUNC_ENDED;
			return NULL;
		}
		if (ck->change) {
			ck->changeCond.unlock();
			continue;
		}
		ck->changeCond.wait(now + sleepTime);
		ck->changeCond.unlock();
	}
}

bool CacheKeeper::setLimits(long minValue, long maxCost)
{
	this->minValue = minValue;
	this->maxCost = maxCost;
	return imposeCostLimit(maxCost);
}

bool CacheKeeper::imposeCostLimit(long maxCost)
{
	while (true) {
		changeCond.lock();
		if (totalCost <= costLimit) {
			changeCond.unlock();
			return true;
		}
		changeCond.unlock();
		mntrLock.lock();
		long sleepTime = 0;
		TimeStamp now = TimeStamp::now();
		bool retval = maintainerStep(now, sleepTime, maxCost);	/* ignore sleepTime */
		mntrLock.unlock();
		if (!retval) {
			return false;
		}
	}
	return true;
}

void CacheKeeper::wakeUpMntr(void* instance, long costChange)
{
	CacheKeeper* ck = (CacheKeeper *)instance;
	if (ck == NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	}
	ck->changeCond.lock();
	ck->change = true;
	ck->totalCost += costChange;
	if (ck->totalCost > ck->costLimit) {
		ck->changeCond.unlock();
		ck->imposeCostLimit(ck->costLimit);
		ck->changeCond.lock();
	}
	ck->changeCond.signal();
	ck->changeCond.unlock();
}

}/*namespace iopc*/
