#ifndef CACHEKEEPER_H_
#define CACHEKEEPER_H_

// project: POLiTe
// authors: Jan Hadrava 2004

/** \file */

#include <map>

#include "composedCache.h"
#include "strategy.h"
#include "simpleStrategySelector.h"
#include "simpleCacheSelector.h"
#include "iopccommon/condVar.h"
#include "iopccommon/rwLock.h"
#include "iopccommon/defs.h"
#include "../db/cachedConnection.h"

namespace iopc {

/** A class making it easy to manage object caches. Can be used to compose caches, set strategies,
run asynchronous maintenance. */
class DLL_EXTERNAL CacheKeeper {
protected:
	/** A record internally representing settings for one connection */
	typedef struct {
		ComposedCache* cache;	/**< The cache assigned to the connection. */
		SimpleCacheSelector* cacheSel; /**< Represents subcaches making up the composed cache. */
		SimpleStrategySelector* stratSel; /**< Represents strategy settings. */
	} CKRecord;
	typedef std::map<CachedConnection*, CKRecord*> CKMap;
	typedef CKMap::iterator Iter;
public:
	/**
	* \param minValue Minimal value of objects in managed caches.
	* \param maxCost Soft limit of total cost of all caches. Maintained asynchronously.
	* \param costLimit Hard limit of total cost of all caches. Maintained synchronously.
	* \param memCost Indicates what type of object cost computation should be used (false - count of items, true - size of memory used by all persistent attributes of all objects in all caches).
	*/
	CacheKeeper(long minValue = MINLONG, long maxCost = MAXLONG, long costLimit = MAXLONG, bool memCost = false);
	~CacheKeeper();
public:
	/** Sets cache for specified connection and object type.
	* \param conn The connection cache should be set for.
	* \param proto the type cache should be set for.
	* \param cache The cache to be set.
	*/
	ComposedCache* setCache(CachedConnection* conn, const Type& type, Cache* cache);
	/** Sets cache for specified connection.
	* \param conn The connection cache should be set for.
	* \param cache The cache to be set.
	*/
	ComposedCache* setCache(CachedConnection* conn, Cache* cache);
	/** Sets strategy for specified connection and object type.
	*/
	ComposedCache* setStrategy(CachedConnection* conn, const Type& type, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat);
	/** Sets strategy for specified connection.
	*/
	ComposedCache* setStrategy(CachedConnection* conn, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat);
	/** Sets strategy for specified connection and object type.
	*/
	ComposedCache* setStrategy(CachedConnection* conn, const Type& type, const Strategy& strat);
	/** Sets strategy for specified connection.
	*/
	ComposedCache* setStrategy(CachedConnection* conn, const Strategy& strat);
	/** Breaks the mutual assignment of connection and respective composed cache.
	* Release internal structures representing settings for the connection. */
	void releaseConnection(CachedConnection* conn);
	/** Wake up maintainer provided by cache keeper. */
	static void wakeUpMntr(void* instance, long costChange);
public:
	/** Run asynchronous maintainer. */
	bool runMaintainer();
	/** Stop asynchronous maintainer.*/
	bool stopMaintainer();
protected:
	ComposedCache* setCacheInternal(CachedConnection* conn, const Type* type, Cache* cache);
	ComposedCache* setStrategyInternal(CachedConnection* conn, const Type* type, bool loadWait, bool loadLocking, UpdatingStrategy uStrat, ReadingStrategy rStrat);

	/** Executes one maintenance step
	* \param sleepTime. Output parameter, the time (in milliseconds) maintainer can sleep for unless it's woken up.
	* \param limit. The total cost to be achieved.
	*/
	bool maintainerStep(const TimeStamp& now, long& sleepTime, long limit);
	/** Main function for asynchronous maintainence thread.
	* \param instance (void *)this
	*/
	static void* maintainer(void *instance);
	/** Resets limits. */
	bool setLimits(long minValue = MINLONG, long maxCost = MAXLONG);
	/** Tries to impose cost limit
	* \return true on success, false on fail.
	*/
	bool imposeCostLimit(long maxCost = MAXLONG);
protected:
	CKMap ckMap;	/** Container holding connection records. */
	bool change;
	CondVar changeCond;
	Mutex mntrLock;
	RWLock lock;
	long maxCost;
	long minValue;
	long costLimit;
	bool memCost;
	long totalCost;
	bool mntrRunning;
	bool stopMntr;
	pthread_t mntrId;
};

}/*namespace iopc*/



#endif /*CACHEKEEPER_H_*/
