#ifndef LRUCACHEST_H_
#define LRUCACHEST_H_

// project: POLiTe
// authors: Jan Hadrava 2004

/** \file */
#include <map>

#include "iopccommon/defs.h"

#ifndef WIN32
#include <ext/hash_map>
#endif /*WIN32*/

#include "lruCacheBase.h"
#include "types.h"

namespace iopc {

template<typename MapType> class LruCacheST;

typedef LruCacheST<std::map<PersIdentification, LruCacheItem* > > MapLruCacheST;

#ifndef WIN32
typedef LruCacheST<__gnu_cxx::hash_map<PersIdentification, LruCacheItem *> > HashLruCacheST;
#endif /*WIN32*/

/** Single-threaded cache using LRU replacement policy. Objects are represented by LruCacheItem.
* Items are stored an container conforming to STL "pair associative container" concept. Its type
* is specified by template parameter.
*/
template<typename MapType> class LruCacheST: public LruCacheBase
{
private:
#ifndef WIN32
	typedef class MapType::iterator Iter;	// MSVC6 doesn't like this although it's perfectly conforming to C++ standard
#else
	typedef MapType::iterator Iter;		// gcc doesn't like this
#endif /*WIN32*/

public:
	inline LruCacheST(long minValue = MINLONG, long maxCost = MAXLONG, const Strategy& strat = Strategy::defaultStrategy);
	inline LruCacheST(long minValue, long maxCost, StrategySelector* stratSel);
	inline virtual ~LruCacheST();
	
private:
	LruCacheST(const LruCacheST<MapType>& src);
	LruCacheST<MapType> &operator=(const LruCacheST<MapType>& );
	
protected:
	virtual void clear(const StrategySelector& strSel);

	virtual CacheLock& makePersistent(const StrategySelector& strSel, const PersIdentification& persId, DatabaseObject& objectPtr);
	virtual CacheLock& getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly = false);

	virtual void dbDelete(const StrategySelector& strSel, const PersIdentification& persId);
	virtual void cacheDelete(const StrategySelector& strSel, const PersIdentification& persId);

	virtual void updateDirty(const StrategySelector& strSel);
	virtual void removeAll(const StrategySelector& strSel);
	
protected:
	LruCacheItem& addLock(LruCacheItem& item, bool rdonly);
	void rmLock(LruCacheItem& item);
	
public:
	virtual inline long getActiveCost();
	virtual inline long getLockedCost();
	virtual long getWorstValue(const TimeStamp& current);
	virtual long removeWorst();
	virtual bool maintainerStep(long& costChange);
	virtual long getSleepTime(const TimeStamp& current, long minValue);
	inline virtual Cache* clone() {
		return new LruCacheST<MapType>(*this);
	}
protected:
	virtual bool removeDoomed();
	virtual bool removeZombie();
	virtual bool asyncUpdate(long& costChange);
protected:
	MapType cont;
	long activeCnt, lockedCnt;
	LruCacheItem activeList;
	LruCacheItem doomedList;	/* double linked list of items scheduled for deletion */
	LruCacheItem zombieList;	/* double linked list of item remnants */
	long activeCost, lockedCost;	/* doomedCost == 0 && zombieCost == 0 */
};

template<typename MapType> LruCacheST<MapType>::LruCacheST(long minValue, long maxCost, StrategySelector* stratSel)
: LruCacheBase(minValue, maxCost, stratSel),
	cont(),
	activeCnt(0), lockedCnt(0),
	activeList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCacheST<MapType>::LruCacheST(long minValue, long maxCost, const Strategy& strat)
: LruCacheBase(minValue, maxCost, strat),
	cont(),
	activeCnt(0), lockedCnt(0),
	activeList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCacheST<MapType>::LruCacheST (const LruCacheST<MapType>& src)
: LruCacheBase(src.minValue, src.maxCost, src.stratSel, src.defaultStrategy),
	cont(),
	activeCnt(0), lockedCnt(0),
	activeList(), doomedList(), zombieList(),
	activeCost(0), lockedCost(0)
{
}

template<typename MapType> LruCacheST<MapType>::~LruCacheST()
{
	ExtendedCache::clear();
}

template<typename MapType> void LruCacheST<MapType>::clear(const StrategySelector& strSel)
{
	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		long cost;
		switch (item->state) {
		case LruCacheItem::ST_BORN:	/* new item */
			if (item->getObject().isDirty()) {
				try {
					persId.insert(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			item->remove();
			activeCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
		break;
		case LruCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				try {
					persId.update(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}
			item->remove();
			activeCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
		break;
		case LruCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			try {
				persId.remove(*conn);
			}
			IOPC_LOG_CATCH(IopcException& exc)
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			item->remove();
		break;
		case LruCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			item->remove();
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}

		delete item;
	}
	cont.clear();

	IOPC_ASSERT(activeCnt == 0 && lockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(activeList.prev == &activeList && activeList.next == &activeList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == &doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == &zombieList);
	const Strategy& strat = strSel.select();
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> CacheLock& LruCacheST<MapType>::makePersistent(const StrategySelector &strSel, const PersIdentification& persId, class DatabaseObject& objectPtr)
{
	long costChange;
	const Strategy& strat = strSel.select(persId);
	LruCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
			case LruCacheItem::ST_BORN:
			case LruCacheItem::ST_ADULT:
				IOPC_LOG_THROW(IopcDuplicateKeyException());
				break;
			default:
				IOPC_UNREACHABLE;
		}
		LruCacheItem::State oldstate = item->state;
		item->state = LruCacheItem::ST_BUSY;	/* we will keep the item in the cache */
		switch (oldstate) {
			case LruCacheItem::ST_BORN:
			case LruCacheItem::ST_ADULT:
				IOPC_UNREACHABLE;
				break;
			case LruCacheItem::ST_DOOMED:
				try {
					persId.remove(*conn);	/* doing maintainer's work */
				}
				IOPC_LOG_CATCH(IopcException& exc)
					item->state = LruCacheItem::ST_DOOMED;
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
				break;
			case LruCacheItem::ST_ZOMBIE:
				break;
			case LruCacheItem::ST_BUSY:
			default:
				IOPC_UNREACHABLE;
			break;
		}
		item->remove();
		/* now we have in cache busy item which is prepared to become born/adult */
	} else {
		item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
		cont[persId] = item;
	}
	IOPC_ASSERT(item->state == LruCacheItem::ST_BUSY);
	UpdatingStrategy uStrat = strat.getUpdatingStrategy();
	if (uStrat == USImmed && strat.getLoadLocking()) {
		uStrat = USDeferred;	/* it's not possible to immediately insert object which should be locked in database */
	}
	if (uStrat == USImmed) {
		objectPtr.setDirty(false);
	} else {
		objectPtr.setDirty(true);
	}
	if (uStrat == USImmed) {
		try {
			persId.insert(*conn, objectPtr);
		}
		IOPC_LOG_CATCH(IopcException& exc)
			zombieList.addPrev(*item);
			costChange = 0;
			item->state = LruCacheItem::ST_ZOMBIE;
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		objectPtr.setDirty(false);
	}
	item->isLocked = true;
	item->addLock(false);
	//objectPtr.setPersistent(*conn);
	item->objectPtr = &objectPtr;

	lockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());

	item->state = uStrat == USImmed ? LruCacheItem::ST_ADULT : LruCacheItem::ST_BORN;

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> CacheLock& LruCacheST<MapType>::getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly)
{
	const Strategy& strat = strSel.select(persId);
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		LruCacheItem* item = iter->second;
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case LruCacheItem::ST_BORN:
		case LruCacheItem::ST_ADULT: {
		/* item is present and active */
			item->addLock(rdonly);
			if (item->isLocked) {
				/* item was already locked,
				** so there's no need to remove it to cleanList
				** and it will be unsafe to reload it */
				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* first lock added, item must be added to lockedList and can be reloaded */
			item->remove(); /* removes item from activeList or updateList */
			activeCnt--;
			activeCost -= strat.cost(item->getObject());
			lockedCnt++;
			lockedCost += strat.cost(item->getObject());
			item->isLocked = true;
			ReadingStrategy rStrat = strat.getReadingStrategy();
			if (item->state == LruCacheItem::ST_BORN || rStrat == RSCache || item->getObject().isDirty()) {
				/* there's no need to reload object */
				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* object will be reloaded */
			IOPC_ASSERT(item->state == LruCacheItem::ST_ADULT);
			item->state = LruCacheItem::ST_BUSY;
			try {
				switch (rStrat) {
				case RSTimestamp:
					if (persId.getIDType() == PersIdentification::OID) {
						if (persId.versionMatches(*conn, dynamic_cast<OidObject&>(item->getObject()))) {
							break;
						}
					} else {
						LogWriter::logWarning("Can't use reading strategy RSTimestamp with non oid-based object. Using RSTimestamp.");
					}
				/*nobreak*/
				case RSDatabase: {
					DatabaseObject* objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
					IOPC_ASSERT(objectPtr != NULL);
					item->replaceObject(objectPtr);
				break; }
				case RSCache:
				break;
				default:
					IOPC_UNREACHABLE;
				break;
				}
			}
			IOPC_LOG_CATCH(IopcException&)	/* reload failed, ignore it */
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			item->state = LruCacheItem::ST_ADULT;
			strat.wakeUpMaintainer(0);
			imposeLimits();
			return *item;
		break; }
		case LruCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */
			IOPC_LOG_THROW(IopcException("Item not present"));
		break;
		case LruCacheItem::ST_ZOMBIE:
			item->remove();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
	}
	IOPC_ASSERT(cont.find(persId) == cont.end());
	LruCacheItem* item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
	cont[persId] = item;
	DatabaseObject* objectPtr;
	try {
		objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
	}
	IOPC_LOG_CATCH(IopcException& exc)
		zombieList.addPrev(*item);
		item->state = LruCacheItem::ST_ZOMBIE;
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	IOPC_ASSERT(objectPtr != NULL);
	//objectPtr->setPersistent(*conn);
	item->objectPtr = objectPtr;
	item->addLock(rdonly);
	item->isLocked = true;
	long costChange;
	lockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());
	item->state = LruCacheItem::ST_ADULT;

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> void LruCacheST<MapType>::dbDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	long costChange = 0;
	const Strategy& strat = strSel.select(persId);
	LruCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
			case LruCacheItem::ST_BORN:	/* remove it immediately */
				item->remove();
				activeCnt--;
				activeCost += costChange = -strat.cost(item->getObject());
				cont.erase(iter);
				delete item;
				strat.wakeUpMaintainer(costChange);
				imposeLimits();
				return; /* ignoring the possibility an item already present in the database
					** can have the same key as the removed one */
			break;
			case LruCacheItem::ST_DOOMED:
			/* item is on the verge of extinction */
				IOPC_LOG_THROW(IopcException("Item not present"));
			break;
			case LruCacheItem::ST_ADULT:
				item->remove();
				activeCnt--;
				activeCost += costChange = -strat.cost(item->getObject());
				item->release();
				item->state = LruCacheItem::ST_BUSY;
			break;
			case LruCacheItem::ST_ZOMBIE:
				item->remove();
				item->state = LruCacheItem::ST_BUSY;
				item->release();	/* release data */
				costChange = 0;
			break;
			case LruCacheItem::ST_BUSY:
			default:
				IOPC_UNREACHABLE;
			break;
			}
		//IOPC_UNREACHABLE;
	} else {
		item = new LruCacheItem(*this, strat, persId, NULL, LruCacheItem::ST_BUSY);
		cont[persId] = item;
		costChange = 0;
	}
	IOPC_ASSERT(item->state == LruCacheItem::ST_BUSY);
	if (strat.getUpdatingStrategy() == USImmed) {
		zombieList.addPrev(*item);
		try {
			persId.remove(*conn);
		}
		IOPC_LOG_CATCH(IopcException& exc)
			item->state = LruCacheItem::ST_ZOMBIE;
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->state = LruCacheItem::ST_ZOMBIE;
	} else { /* USAsync || USImmedDeferred */
		doomedList.addPrev(*item);	/* schedule for deletion */
		item->state = LruCacheItem::ST_DOOMED;
	}
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void LruCacheST<MapType>::cacheDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	/* ignoring user's advices */
}

template<typename MapType> void LruCacheST<MapType>::updateDirty(const StrategySelector& strSel)
{
	Iter next;
	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter = next) {
		next = iter;
		next++;
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case LruCacheItem::ST_BORN:	/* new item */
			IOPC_ASSERT(item->getObject().isDirty());
			persId.insert(*conn, item->getObject());
			item->getObject().setDirty(false);
			item->state = LruCacheItem::ST_ADULT;
		break;
		case LruCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				persId.update(*conn, item->getObject());
				item->getObject().setDirty(false);
			}
		break;
		case LruCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			persId.remove(*conn);
			item->remove();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */
			item->remove();
			cont.erase(iter);
			delete item;
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		if ((item->state == LruCacheItem::ST_BORN || item->state == LruCacheItem::ST_ADULT)
			&& strat.getLoadLocking())
		{	/* item with locking must be removed */
			item->remove();
			activeCnt--;
			long cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;
			delete item;
			cont.erase(iter);
		}
	}
	IOPC_ASSERT(doomedList.next == &doomedList && doomedList.prev == &doomedList);
	IOPC_ASSERT(zombieList.next == &zombieList && zombieList.prev == &zombieList);

	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void LruCacheST<MapType>::removeAll(const StrategySelector& strSel)
{
	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		LruCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */
		IOPC_ASSERT(item->state != LruCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		item->remove();
		switch (item->state) {
		case LruCacheItem::ST_BORN:
		case LruCacheItem::ST_ADULT:
			activeCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
		break;
		case LruCacheItem::ST_DOOMED:
		case LruCacheItem::ST_ZOMBIE:
		break;
		case LruCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		delete item;
	}
	cont.clear();
	IOPC_ASSERT(activeCnt == 0 && lockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(activeList.prev == &activeList && activeList.next == & activeList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == & doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == & zombieList);

	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> LruCacheItem &LruCacheST<MapType>::addLock(LruCacheItem& item, bool rdonly)
{
	if (rdonly) {
		item.lock.rlock();
	} else {
		item.lock.wlock();
	}
	return item;
}

template<typename MapType> void LruCacheST<MapType>::rmLock(LruCacheItem& item)
{
	/* it's important not to use contLock here so that we avoid deadlock */
	try {
		item.lock.unlock();
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	if (!item.isLocked) {
		return;
	}
	if (!item.lock.tryXLock()) {
		return;
	}
	item.remove();	/* remove from list of locked items */
	long cost = item.strat.cost(item.getObject());
	long costChange = 0;
	lockedCnt--;
	lockedCost -= cost;
	item.isLocked = false;
	activeCnt++;
	activeCost += cost;
	UpdatingStrategy uStrat = item.strat.getUpdatingStrategy();
	if (uStrat == USImmed && item.state == LruCacheItem::ST_BORN) {
		IOPC_ASSERT(item.strat.getLoadLocking());
		/* this combination will require inserting item into db and removing local copy but we can't do it
		because it requires state change which can't be made without contLock */
		uStrat = USAsync;
	}
	switch (uStrat) {
	case USDeferred:
		activeList.addNext(item);
	break;
	case USImmed:
		if (item.state == LruCacheItem::ST_BORN) {
			IOPC_UNREACHABLE;
		} else {
			IOPC_ASSERT(item.state == LruCacheItem::ST_ADULT);
			if (item.getObject().isDirty()) {
				item.persId.update(*conn, item.getObject());
				item.getObject().setDirty(false);
			}
			activeList.addNext(item);
		}
	break;
	case USAsync:
		activeList.addNext(item);
	break;
	}
	item.timeStamp = TimeStamp::now();
	Strategy strat = item.strat;
	item.lock.unlock();
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> inline long LruCacheST<MapType>::getActiveCost()
{
	return activeCost;
}

template<typename MapType> inline long LruCacheST<MapType>::getLockedCost()
{
	return lockedCost;
}

template<typename MapType> long LruCacheST<MapType>::getWorstValue(const TimeStamp& time)
{
	long retval = MAXLONG;
	if (activeList.prev != &activeList) {
		retval = activeList.prev->timeStamp - time;	/* difference in milliseconds */
	}
	return retval;
}

template<typename MapType> long LruCacheST<MapType>::removeWorst()
{
	//LruCacheItem* list = NULL;
	long value;
	TimeStamp now = TimeStamp::now();
	LruCacheItem* item = NULL;
	if (activeList.prev != &activeList) {
		value = activeList.prev->timeStamp - now;
		item = activeList.prev;
	}
	if (item == NULL) {
		return 0;
	}
	if (item->state == LruCacheItem::ST_BUSY) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("Incorrect behavior of maintainer"));
	}
	bool isAdult;
	if (item->state == LruCacheItem::ST_ADULT) {
		isAdult = true;
	} else {
		IOPC_ASSERT(item->state == LruCacheItem::ST_BORN);
		isAdult = false;
	}
	item->state = LruCacheItem::ST_BUSY;
	item->remove();
	long cost;
	activeCnt--;
	activeCost += cost = -item->strat.cost(item->getObject());
	zombieList.addPrev(*item);
	if (item->getObject().isDirty()) {
		if (isAdult) {
			item->persId.update(*conn, item->getObject());
		} else {
			item->persId.insert(*conn, item->getObject());
		}
		item->getObject().setDirty(false);
	}
	item->release();
	item->state = LruCacheItem::ST_ZOMBIE;
	return cost;
}

template<typename MapType> bool LruCacheST<MapType>::removeDoomed()
{
	if (doomedList.prev == &doomedList) {
		return false;
	}
	LruCacheItem &item = *doomedList.prev;
	IOPC_ASSERT(item.state == LruCacheItem::ST_DOOMED);
	item.state = LruCacheItem::ST_BUSY;
	item.remove();
	zombieList.addPrev(item);
	item.persId.remove(getConnection());
	item.state = LruCacheItem::ST_ZOMBIE;
	return true;
}

template<typename MapType> bool LruCacheST<MapType>::removeZombie()
{
	if (zombieList.next == &zombieList) {
		return false;
	}
	LruCacheItem *item = zombieList.next;
	IOPC_ASSERT(item->state == LruCacheItem::ST_ZOMBIE);
	Iter iter = cont.find(item->persId);
	IOPC_ASSERT(iter != cont.end());
	cont.erase(iter);
	item->remove();
	delete item;
	return true;
}

template<typename MapType> bool LruCacheST<MapType>::asyncUpdate(long& costChange)
{
	costChange = 0;
	return false;
}

template<typename MapType> bool LruCacheST<MapType>::maintainerStep(long& costChange)
{
	costChange = 0;
	bool retval = asyncUpdate(costChange);
	if (removeDoomed()) {
		retval = true;
	}
	if (removeZombie()) {
		retval = true;
	}
	return retval;
}

template<typename MapType> long LruCacheST<MapType>::getSleepTime(const TimeStamp& current, long minValue)
{
	LruCacheItem *item = NULL;
	long diff;

	if (activeList.prev != &activeList) {
		diff = activeList.prev->timeStamp - current;
		item = activeList.prev;
	}
	if (item == NULL) {
///		return 1800*1000000;	/* something like infinity */
		return 5*1000000;
	}
	if (diff < minValue) {
		return 0;		/* no time to sleep */
	}
	return diff - minValue;
}

}/*namespace iopc*/

#endif /*LRUCACHEST_H_*/
