#include <map>
#include "voidCache.h"
#include "voidCacheItem.h"
#include "iopcdb/connection.h"
#include "iopccommon/exceptionBase.h"
#include "iopccommon/trace.h"
#include <string>

using namespace std;

namespace iopc {

VoidCache::VoidCache()
{
	IOPC_FUNC_INVOKED;
	items = 0;
}

VoidCache::~VoidCache()
{
	IOPC_FUNC_INVOKED;
}

VoidCache::VoidCache(const VoidCache& cache)
{
	IOPC_FUNC_INVOKED;
	items = 0;
}

Cache* VoidCache::clone()
{
	IOPC_FUNC_INVOKED;
	return new VoidCache(*this);
}

void VoidCache::clear()
{
	IOPC_FUNC_INVOKED;
	emptyCond.lock();
	if (items == 0) {
		emptyCond.unlock();
		return;
	}
	emptyCond.unlock();
	
	contLock.wlock();

	bool locked;
	bool err;
	bool errMsg = false;
	for (IterType iter = cont.begin(); iter != cont.end(); iter++) {
		locked = false;
		err = false;
		VoidCacheItem *item = iter->second;
		try {
			if (item->isItemLocked()) {
				locked = true;
				contLock.unlock();
				return;
			}
		}
		IOPC_LOG_CATCH(IopcException& exc)
			LogWriter::logError(string("Void Cache :") + exc.what());
			err = true;
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		
		if (!err || !locked) {
			delete item;	/* this will wait for all locks to be unlocked */
		} else {
			errMsg = true;
		}
	}
	cont.clear();
	emptyCond.lock();
	items = 0;
	emptyCond.signal();
	emptyCond.unlock();
	contLock.unlock();

	LogWriter::logWarning("There were existing cached items found when clearing the VoidCache. This shouldn't happen because the VoidCache contains only locked items.");
	
	if (errMsg) {
		LogWriter::logError("There were problems deleting cached items from the VoidCache. Problems are marked \"Void Cache:\" above in the log. Some items remained in memory.");
	}
}

CacheLock& VoidCache::makePersistent(const PersIdentification& persId, DatabaseObject& objectPtr)
{
	IOPC_FUNC_INVOKED;
	contLock.wlock();
	IterType iter = cont.find(persId);
	if (iter != cont.end()) {
		contLock.unlock();
		IOPC_LOG_THROW(IopcDuplicateKeyException());
	}
	//objectPtr.setPersistent(getConnection());
	objectPtr.setDirty(true);
	VoidCacheItem* item = new VoidCacheItem(*this, persId, objectPtr, true);
	item->addLock(false);
	cont[persId] = item;
	emptyCond.lock();
	items++;
	emptyCond.unlock();
	contLock.unlock();
	return *item;
}

CacheLock& VoidCache::getCacheLock(const PersIdentification& persId, bool rdonly)
{
	IOPC_FUNC_INVOKED;
	contLock.wlock();
	IterType iter = cont.find(persId);
	VoidCacheItem* item;
	if (iter != cont.end()) {
		item = iter->second;
	} else {
		DatabaseObject* objectPtr;
		try {
			objectPtr = persId.load(getConnection(), true, true);
			IOPC_ASSERT(objectPtr != NULL);
			//objectPtr->setPersistent(getConnection());
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item = new VoidCacheItem(*this, persId, *objectPtr, false);
		cont[persId] = item;
		emptyCond.lock();
		items++;
		emptyCond.unlock();
	}
	try {
		item->addLock(rdonly);
	}
	IOPC_LOG_CATCH(IopcException& exc)
		contLock.unlock();
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	contLock.unlock();
	return *item;
}

void VoidCache::dbDelete(const PersIdentification& persId)
{
	IOPC_FUNC_INVOKED;
	contLock.wlock();
	IterType iter = cont.find(persId);
	if (iter != cont.end()) {
		VoidCacheItem* item = iter->second;
		item->unlock();
		cont.erase(iter);
		delete item;	/* this will wait for all locks to be unlocked */
		IOPC_ASSERT(items > 0);
		emptyCond.lock();
		items--;
		emptyCond.signal();
		emptyCond.unlock();
	}
	try {
		persId.remove(getConnection());
	}
	IOPC_LOG_CATCH(IopcException& exc)
		contLock.unlock();
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	contLock.unlock();
}

void VoidCache::cacheDelete(const PersIdentification &persId)
{
	IOPC_FUNC_INVOKED;
/*there's no need to do anything, because this implementation of cache can contain
**only locked items */
}

/* called when unlocking an item */
void VoidCache::tryCacheDelete(PersIdentification &persId)
{
	IOPC_FUNC_INVOKED;
	contLock.wlock();
	IterType iter = cont.find(persId);
	if (iter != cont.end()) {
		VoidCacheItem *item = iter->second;
		
		try {
			if (item->isItemLocked()) {
				contLock.unlock();
				return;
			}
		}
		IOPC_LOG_CATCH(IopcException& exc)
			contLock.unlock();
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		
		if (item->getObject().isDirty()) {
			try {
				if (item->isNew()) {
					persId.insert(getConnection(), item->getObject());
				} else {
					persId.update(getConnection(), item->getObject());
				}
			}
			IOPC_LOG_CATCH(IopcException& exc)
				contLock.unlock();
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
			
		}
		
		cont.erase(iter);
		delete item;	/* this will wait for all locks to be unlocked */
		
		emptyCond.lock();
		IOPC_ASSERT(items > 0);
		items--;
		emptyCond.signal();
		emptyCond.unlock();
	}
	contLock.unlock();
	IOPC_FUNC_ENDED;
}

void VoidCache::updateDirty()
{
	IOPC_FUNC_INVOKED;
	emptyCond.lock();
	while (items > 0) {
		emptyCond.wait();
	}
	emptyCond.unlock();
	IOPC_ASSERT(cont.begin() == cont.end());
}

void VoidCache::removeAll()
{
	IOPC_FUNC_INVOKED;
	clear();
}

}
