#ifndef DBPTRBASE_H_
#define DBPTRBASE_H_

#include "iopccommon/defs.h"
#include "iopccommon/exceptionBase.h"
#include "db/cachedConnection.h"
#include "cache/cacheLock.h"
#include "iopcmeta/object.h"
#include "persIdentification.h"
#include "types/dbType.h"
#include "keyValues.h"
#include "iopccommon/countedPtr.h"
#include <string>

namespace iopc {

/**
 * Class representing non-typed database pointer.
 */
class DLL_EXTERNAL DbPtrBase : public EnhancedTypeBase {
public:

	enum DbPtrState {TRANSIENT = 0, PERSISTENT = 1};

	DbPtrBase(const Type& type/*, bool createInstance = true*/)
		throw ();

	/*DbPtrBase(const Type& type, CachedConnection* conn)
		throw ();*/

	DbPtrBase(const Type& type, CachedConnection* conn, const OidType& oid, bool loadRealType = true)
		throw ();

	DbPtrBase(const Type& type, CachedConnection* conn, const KeyValues& keyVals)
		throw ();

	DbPtrBase(const DbPtrBase& src)
		throw ();

	virtual ~DbPtrBase()
		throw ();

private: /*hiding*/
	DbPtrBase();
	DbPtrBase& operator=(const DbPtrBase& src);

	DbPtrState state;

public:
	/** Inspects database pointer validity. */
	operator bool() const
		throw ();

	/** Makes the referenced object persistent. */
	CacheLock& bePersistent(CachedConnection* conn);

	/** Removes referenced object from database */
	void dbDelete()
		throw (IopcIncorrectUsageException);

	/** Hints cache to drop its copy of referenced object. */
	void cacheDelete() const
		throw (IopcIncorrectUsageException);

	/** Creates buffer containing all key values. */
	/*char *SelectKeyValues() const
		throw ();*/

	/** Inspects database pointer validity. */
	inline bool isValid() const
		throw ();

	/** Inspects persistency/transiency of referenced object. */
	inline bool isPersistent() const
		throw ();

protected:
/*	virtual void setPersId()
		throw (IopcIncorrectUsageException);*/

	//virtual void createInvalidCopyOf(const DbPtrBase& src);

	/**
	 * Replaces content of this instance with the content of the src.
	 *
	 * Called from the = operator
	 */
	virtual void replaceWith(const DbPtrBase& src);

	//void setNewInstance(DatabaseObject* obj);

	virtual void deleteObject();

	inline void callOnIopcCreate(DatabaseObject* o) { o->onIopcCreate(); }

	virtual bool isObjectPtrNull() const =0;
	virtual bool isObjectPtrUnique() const =0;
	virtual DatabaseObject& getObject() const =0;
	virtual void setObjectPtrNull(bool deleteObject = true) =0;

	void propagateConnectionToAttributes(DatabaseObject& o) const;
public:
	inline const PersIdentification& getPersId() const { return persId; }

	inline const DbPtrState getState() const { return state; }

	/**
	 * Connection can be get only for persistent objects
	 */
	CachedConnection& getConnection() const
		throw (IopcInvalidStateException);

	//inline CountedPtr<DatabaseObject> getObjectPtr() const { return objectPtr; }


protected:
	/** Referenced object (used for transient objects). */
	//CountedPtr<DatabaseObject> objectPtr;

	/**
	 * The connection used for manipulation with persistent copy of referenced object.
	 */
	CachedConnection* conn;

	/** ID of referenced instance (used for persistent objects). */
	PersIdentification persId;

	inline void setConnection(CachedConnection* conn) { this->conn = conn; }

	/**********  Enhanced type base *************/

private:
	OidType oidCopy;

public:
	virtual void* externalizeCopy(unsigned int maxlength = 0);

	virtual void* externalizeEmptyBuffer(unsigned int length = 0);

	virtual void updateFromExternalized();

	virtual unsigned int getBufferSize();

	virtual const Type& getInnerType() const;

	//void checkExternalizationOperation();

	bool updateValue(OidType newVal);

	/**********  Objecct *************/
protected:
	virtual const Type& getTypeInternal() const;

};

inline bool DbPtrBase::isPersistent() const
	throw ()
{
	//IOPC_ASSERT(persId.proto != NULL);
	return state == PERSISTENT;
}

inline bool DbPtrBase::isValid() const
	throw ()
{
	//IOPC_ASSERT(persId.proto != NULL);
	return conn != NULL || !isObjectPtrNull();
}


template <> class DLL_EXTERNAL TypeDesc<DbPtrBase> : public Type {
public:
	inline virtual DataTypeEnum getDataType() const { return IOPC_TYPE_REF; }
	inline virtual const std::string getName() const { return "DbPtrBase"; }
	inline virtual const std::string getNamespace() const { return "iopc"; }
	inline virtual void* newInstance() const { return NULL; }
	inline virtual Object* newObjectInstance() const { return NULL; }
	inline virtual int getSizeOf(const void* val) const { return TypeDesc<OidType>::getType().getSizeOf(val); }
	static inline const TypeDesc<DbPtrBase>& getType() { return instance; }
	inline virtual DataTypeClassEnum getDataTypeClass() const { return IOPC_TYPECLASS_ENHANCED; }
	inline virtual void fillAttributeValues(const Object& o, std::map<std::string, AttributeValue>& out) const {}
	inline virtual bool isBuiltIn() const { return true; }
private:
	TypeDesc() { TypeManager::registerType(this, "iopc::DbPtrBase", &typeid(DbPtrBase)); }
	static TypeDesc<DbPtrBase> instance;
};

}



#endif /*DBPTRBASE_H_*/
