#include "orMappingStrategy.h"
#include "dbType.h"
#include "dbTypeManager.h"
#include "iopcdb/connection.h"
#include "../db/orStatementsFeature.h"
#include "iopcdb/cursor.h"
#include "iopcdb/driver.h"

using namespace iopc;
using namespace std;

void ORMappingStrategy::init()
{
	hasFilteredColumns = false;
	string tmp = getType().getMeta("db.mapping.type").getStringValue();
	if (getDbType().getMappingType() == DbType::MAPPING_FILTERED) {

		if (!getDbType().isOidManaged()) {
			IOPC_LOG_THROW(IopcMetadataException("db.mapping.type", getType().getQualifiedName(), "Filtered mapping is unsupported for non oid-managed types"));
		}

		// db.mapping.insertto attribute is obligatory
		if (!getType().hasMeta("db.mapping.insertto")) {
			IOPC_LOG_THROW(IopcMetadataException("db.mapping.insertto", getType().getQualifiedName(), "Metadata expected for filtered mapping"));
		}

		// insertToType type must exist and must have matching DbType
		try {
			insertToType = &DbTypeManager::getDbTypeWithInit(TypeManager::getType(getType().getMeta("db.mapping.insertto")));
		} IOPC_LOG_CATCH(IopcNotFoundException& e)
			IOPC_LOG_THROW(IopcMetadataException("db.mapping.insertto", getType().getQualifiedName(), (string)"No such database type found: \"" + (string) getType().getMeta("db.mapping.insertto") + "\""));
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		// insertToType must be ancestor of this type
		if (!getType().isTypeOf(insertToType->getType()) || getType() == insertToType->getType()) {
			IOPC_LOG_THROW(IopcMetadataException("db.mapping.insertto", getType().getQualifiedName(), (string)"Insert to type \"" + (string) getType().getMeta("db.mapping.insertto") + "\" is not ancestor of this type"));
		}

		// insertToType can't have MAPPING_FILTERED or MAPPING_OBJECT mapping
		if (insertToType->getMappingType() == DbType::MAPPING_FILTERED || insertToType->getMappingType() == DbType::MAPPING_OBJECT) {
			IOPC_LOG_THROW(IopcMetadataException("db.mapping.insertto", getType().getQualifiedName(), (string)"Insert to type \"" + (string) getType().getMeta("db.mapping.insertto") + "\" can't have filtered or object mapping"));
		}

		if (getDbType().getPersistentAttributes().size() > 0) {
			((ORMappingStrategy&)(insertToType->getMappingStrategy())).addFilteredType(getDbType());
		}
	}

	if (getDbType().getMappingType() == DbType::MAPPING_FILTERED && getDbType().getAllParentsSet().find(insertToType) == getDbType().getAllParentsSet().end()) {
		IOPC_LOG_THROW(IopcMetadataException("db.mapping.insertto", getType().getQualifiedName(), "The target type can't be ancestor of horizontally mapped type. All paths in the type hierarchy from this type to the type specified by insertto metadata contain a horizontally mapped type."));
	}

	if (getDbType().getMappingType() == DbType::MAPPING_FILTERED) {
		hasTable = false;
	} else {
		hasTable = true;
	}

	/*if (allParents.size() == 0) {
		mapping = MAPPING_HORIZONTAL;
	}*/
}

void ORMappingStrategy::postInit()
{
	IOPC_DEBUGLOG(string("[class]: ") + getType().getQualifiedName());
	// Create mappedAttributes vector
	if (hasTable) {
		for (DbType::PersistentAttributesIterator it = getDbType().getPersistentAttributes().begin(); it != getDbType().getPersistentAttributes().end(); it++) {
			IOPC_DEBUGLOG(string("[mappedAttributes - persistent]: ") + (*it)->getName());
			mappedAttributes.push_back(std::make_pair(*it, COLSOURCE_ATTR));
		}

		if (getDbType().getMappingType() == DbType::MAPPING_HORIZONTAL) {
			for (DbType::PersistentAttributesIterator it = getDbType().getInheritedAttributes().begin(); it != getDbType().getInheritedAttributes().end(); it++) {
				if (getDbType().getKeyAttributes().find(*it) == getDbType().getKeyAttributes().end() ) {
					if ((*it)->getOwner().is<OidObject>()) continue;
					IOPC_DEBUGLOG(string("[mappedAttributes - inherited]: ") + (*it)->getName());
					mappedAttributes.push_back(std::make_pair(*it, COLSOURCE_INHERITED));
				}
			}
		}

		if (hasFilteredColumns) {
			for (FilteredTypesIterator it = filteredTypes.begin(); it != filteredTypes.end(); it++) {
				for (DbType::PersistentAttributesIterator ait = (*it)->getPersistentAttributes().begin(); ait != (*it)->getPersistentAttributes().end(); ait++) {
					IOPC_DEBUGLOG(string("[mappedAttributes - filtered]: ") + (*ait)->getName());
					mappedAttributes.push_back(std::make_pair(*ait, COLSOURCE_FILTERED));
				}
			}
		}

		for (DbType::KeyAttributesIterator it = getDbType().getKeyAttributes().begin(); it != getDbType().getKeyAttributes().end(); it++) {
			if ((*it)->getOwner() != getType()) {
				IOPC_DEBUGLOG(string("[mappedAttributes - inheritedpk]: ") + (*it)->getName());
				mappedAttributes.push_back(std::make_pair(*it, COLSOURCE_INHERITEDPK));
			}
		}
	}
}

void ORMappingStrategy::addFilteredType(DbType& dbType)
{
	filteredTypes.insert(&dbType);
	hasFilteredColumns = true;
	IOPC_DEBUGLOG(string("[filteredTypes]: ") + getType().getQualifiedName() + " += " + dbType.getType().getQualifiedName());
}

void ORMappingStrategy::loadOne(Connection& conn, DatabaseObject* obj, const KeyValues& keyVals, bool wait, bool lock)
{
	if (!hasTable) return;

	MetadataHolder meta;
	if (lock) {
		meta.setFeature("db.select.lock");
	}
	if (!wait) {
		meta.setFeature("db.select.nowait");
	}
	const string selectStatement = conn.getDriver().getFeature<ORStatementsFeature>().getSelectFromClassTableStatement(*this, meta);

	Cursor* cur = conn.sql(selectStatement);

	mapKeyAttributes(cur, 1, keyVals);
	int i = 1;
	for (MappedAttributesIterator it = mappedAttributes.begin(); it != mappedAttributes.end(); it++) {

		switch (it->second) {
			case COLSOURCE_FILTERED:
				{
					if (obj->getType().isTypeOf(it->first->getOwner())) {
						const AttributeValue& attrVal = obj->getAttributeValue(it->first->getName());
						cur->addOutParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
					} else {
						//cur->addNullInParam(i, it->first->getType());
					}
				}
				break;
			case COLSOURCE_ATTR:
			case COLSOURCE_INHERITED:
			case COLSOURCE_INHERITEDPK:
				{
					const AttributeValue& attrVal = obj->getAttributeValue(it->first->getName());
					cur->addOutParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
				}
				break;
		}
		i++;
	}

	cur->execute();
	if (!cur->fetchNext()) {
		IOPC_LOG_THROW(IopcNoRowsFetchedException("No row was fetched during the O/R mapping"));
	}
	cur->close();
	conn.returnCursor(cur);
}

OidObject* ORMappingStrategy::viewLoadAll(Connection& conn, const OidType& oid, bool wait, bool lock)
{
	const string selectStatement = conn.getDriver().getFeature<ORStatementsFeature>().getSelectFromClassViewStatement(*this);

	// TODO: udelat dynamic_cast
	OidObject* obj = (OidObject*)getType().newObjectInstance();

	try {
		//LogWriter::logSql(selectStatement);
		Cursor* cur = conn.sql(selectStatement);
		cur->addInParam(1, &oid, TypeDesc<OidType>::getType());
		int i = 1;
		for (DbType::PersistentAttributesIterator it = getDbType().getAllPersistentAttributes().begin(); it != getDbType().getAllPersistentAttributes().end(); it++, i++) {
			const Attribute& attr = *(*it);
			const AttributeValue& attrVal = obj->getAttributeValue(attr.getName());
			//cur->addOutParam(i, )
			cur->addOutParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
		}

		cur->execute();
		//cur->preFetchBind(1, &oid, sizeof(OidType), TypeDesc<OidType>::getType());
		if (!cur->fetchNext()) {
			IOPC_LOG_THROW(IopcNoRowsFetchedException("No row was fetched during the O/R mapping"));
		}
		cur->close();
		conn.returnCursor(cur);
		return obj;

	} catch (...) {
		delete obj;
		throw ;
	}
}

void ORMappingStrategy::updateOne(Connection& conn, DatabaseObject& obj)
{
	Cursor* cur;
	if (!hasTable) return;

	if (!hasFilteredColumns) {
		const string updateStatement = conn.getDriver().getFeature<ORStatementsFeature>().getUpdateStatement(*this);
		cur = conn.sql(updateStatement);
	} else {
		MappedAttributesList updateCols;
		// we can't use pregenerated update statement because
		// the object obj doesn't contains all alltributes that are mapped to columns in this table
		for (MappedAttributesIterator it = mappedAttributes.begin(); it != mappedAttributes.end(); it++)
		{
			if (it->second != COLSOURCE_FILTERED || obj.getType().isTypeOf(it->first->getOwner())) {
				updateCols.push_back(*it);
			}
		}
		const string updateStatement = conn.getDriver().getFeature<ORStatementsFeature>().getUpdateStatement(*this, updateCols);
		cur = conn.sql(updateStatement);
	}

	int i = 1;
	for (MappedAttributesIterator it = mappedAttributes.begin(); it != mappedAttributes.end(); it++)
	{

		switch (it->second) {
			case COLSOURCE_FILTERED:
				{
					if (obj.getType().isTypeOf(it->first->getOwner())) {
						const AttributeValue& attrVal = obj.getAttributeValue(it->first->getName());
						cur->addInParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
						i++;
					} else {
						// Filtered attributes from other type in the hierarchy --> do not update
						//cur->addNullInParam(i, it->first->getType());
					}
				}
				break;
			case COLSOURCE_ATTR:
			case COLSOURCE_INHERITED:
				{
					const AttributeValue& attrVal = obj.getAttributeValue(it->first->getName());
					cur->addInParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
					i++;
				}
				break;
			case COLSOURCE_INHERITEDPK:
				//cur->addInParam(i, &obj.oid, TypeDesc<OidType>::getType());
				break;
		}
	}

	mapKeyAttributes(cur, i, obj);

	cur->execute();
	cur->close();
	conn.returnCursor(cur);
}


void ORMappingStrategy::deleteOne(Connection& conn, const OidType& oid)
{
	IOPC_FUNC_INVOKED
	if (!hasTable) return;
	const string deleteStatement = conn.getDriver().getFeature<ORStatementsFeature>().getDeleteStatement(*this);
	Cursor* cur = conn.sql(deleteStatement);
	mapKeyAttributes(cur, 1, oid);
	cur->execute();
	cur->close();
	conn.returnCursor(cur);
	IOPC_FUNC_ENDED;
}

void ORMappingStrategy::deleteOne(Connection& conn, const KeyValues& keyVals)
{
	IOPC_FUNC_INVOKED
	if (!hasTable) return;
	const string deleteStatement = conn.getDriver().getFeature<ORStatementsFeature>().getDeleteStatement(*this);
	Cursor* cur = conn.sql(deleteStatement);
	mapKeyAttributes(cur, 1, keyVals);
	cur->execute();
	cur->close();
	conn.returnCursor(cur);
	IOPC_FUNC_ENDED;
}

void ORMappingStrategy::insertOne(Connection& conn, DatabaseObject& obj)
{
	IOPC_FUNC_INVOKED;
	if (!hasTable) return;
	const string insertStatement = conn.getDriver().getFeature<ORStatementsFeature>().getInsertStatement(*this);

	Cursor* cur = conn.sql(insertStatement);

	int i = 1;
	for (MappedAttributesIterator it = mappedAttributes.begin(); it != mappedAttributes.end(); it++, i++)
	{
		switch (it->second) {
			case COLSOURCE_FILTERED:
				{
					if (obj.getType().isTypeOf(it->first->getOwner())) {
						const AttributeValue& attrVal = obj.getAttributeValue(it->first->getName());
						cur->addInParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
					} else {
						// Filtered attributes from other type in the hierarchy --> bind NULL
						cur->addNullInParam(i, it->first->getType());
					}
				}
				break;
			case COLSOURCE_ATTR:
			case COLSOURCE_INHERITED:
			case COLSOURCE_INHERITEDPK:
				{
					const AttributeValue& attrVal = obj.getAttributeValue(it->first->getName());
					cur->addInParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
				}
				break;
			/*case COLSOURCE_INHERITEDPK:
				cur->addInParam(i, &obj.oid, TypeDesc<OidType>::getType());
				break;*/
		}
	}

	cur->execute();
	cur->close();
	conn.returnCursor(cur);

	IOPC_FUNC_ENDED;
}
